from pathlib import Path
import re
from setuptools import setup


here = Path(__file__).absolute().parent

version_content = Path(here, "pyrasgo", "version.py").read_text(encoding="utf-8")
version_attrs = dict(re.findall(r"__([a-z]+)__ *= *['\"](.+)['\"]", version_content))
long_description = Path(here / "DESCRIPTION.md").read_text(encoding="utf-8")
requirements = Path(here / "requirements.txt").read_text(encoding="utf-8").splitlines()
sf_requirements = Path(here / "requirements-snowflake.txt").read_text(encoding="utf-8").splitlines()
bq_requirements = Path(here / "requirements-bigquery.txt").read_text(encoding="utf-8").splitlines()
setup(
    name="pyrasgo",
    version=version_attrs["version"],
    description="Python interface to the Rasgo API.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Patrick Dougherty",
    author_email="patrick@rasgoml.com",
    url="https://www.rasgoml.com/",
    license="GNU Affero General Public License v3 or later (AGPLv3+)",
    project_urls={
        "Documentation": "https://docs.rasgoml.com",
        "Source": "https://github.com/rasgointelligence/RasgoSDKPython",
        "Rasgo": "https://www.rasgoml.com/",
        "Changelog": "https://github.com/rasgointelligence/RasgoSDKPython/blob/master/pyrasgo/CHANGELOG.md",
    },
    packages=[
        "pyrasgo",
        "pyrasgo.api",
        "pyrasgo.primitives",
        "pyrasgo.schemas",
        "pyrasgo.storage",
        "pyrasgo.storage.dataframe",
        "pyrasgo.storage.datawarehouse",
        "pyrasgo.utils",
    ],
    install_requires=requirements,
    extras_require={
        "snowflake": sf_requirements,
        "bigquery": bq_requirements,
    },
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
)



"""
DataWarehouse connection functions
"""
from typing import Union

from pyrasgo import errors
from .bigquery import BigQueryDataWarehouse
from .snowflake import SnowflakeDataWarehouse

__all__ = ["dw_connection", "DataWarehouse"]

DataWarehouse = Union[BigQueryDataWarehouse, SnowflakeDataWarehouse]


def dw_connection(organization_dw_type: str) -> DataWarehouse:
    """
    Returns a DataWarehouse class based on the organization
    """
    if organization_dw_type == "BIGQUERY":
        return BigQueryDataWarehouse()
    elif organization_dw_type == "SNOWFLAKE":
        return SnowflakeDataWarehouse()
    else:
        raise errors.DWCredentialsWarning(
            f"{organization_dw_type} is an unrecognized Cloud Data Warehouse. " " Cannot create a connection. "
        )



from datetime import datetime, timedelta
from enum import Enum
from functools import total_ordering
from typing import List, Optional, Dict

from pydantic import BaseModel, Field


class Filter(BaseModel):
    column_name: str = Field(alias='columnName')
    operator: str
    comparison_value: str = Field(alias='comparisonValue')

    class Config:
        allow_population_by_field_name = True


@total_ordering
class TimeGrain(str, Enum):
    HOUR = "HOUR"
    DAY = "DAY"
    WEEK = "WEEK"
    MONTH = "MONTH"
    QUARTER = "QUARTER"
    YEAR = "YEAR"

    def timedelta(self):
        if self.value == "HOUR":
            return timedelta(hours=1)
        if self.value == "DAY":
            return timedelta(days=1)
        if self.value == "WEEK":
            return timedelta(weeks=1)
        if self.value == "MONTH":
            return timedelta(days=30)
        if self.value == "QUARTER":
            return timedelta(days=91)
        if self.value == "YEAR":
            return timedelta(days=365)

    def __lt__(self, other):
        order = (TimeGrain.HOUR, TimeGrain.DAY, TimeGrain.WEEK, TimeGrain.MONTH, TimeGrain.QUARTER, TimeGrain.YEAR)
        if self.__class__ == other.__class__:
            return order.index(self) < order.index(other)
        raise NotImplemented

    def __gt__(self, other):
        order = (TimeGrain.HOUR, TimeGrain.DAY, TimeGrain.WEEK, TimeGrain.MONTH, TimeGrain.QUARTER, TimeGrain.YEAR)
        if self.__class__ == other.__class__:
            return order.index(self) > order.index(other)
        raise NotImplemented


class Metric(BaseModel):
    id: int
    name: str
    ds_dataset_id: int = Field(alias='datasetId')
    aggregation_type: str = Field(alias='aggregationType')
    target_expression: str = Field(alias='targetExpression')
    time_grains: List[str] = Field(alias='timeGrains')
    time_dimension: str = Field(alias='timeDimension')
    dimensions: List[str]

    filters: Optional[List[Filter]]
    meta: Optional[Dict[str, str]]
    label: Optional[str]
    description: Optional[str]

    recent_values: Optional[List] = Field(alias='recentValues')
    recent_time_grain: Optional[str] = Field(alias='recentTimeGrain')
    recent_period_start: Optional[datetime] = Field(alias='recentPeriodStart')
    recent_period_end: Optional[datetime] = Field(alias='recentPeriodEnd')

    class Config:
        allow_population_by_field_name = True


class MetricCreate(BaseModel):
    name: str
    ds_dataset_id: int = Field(alias='datasetId')
    aggregation_type: str = Field(alias='aggregationType')
    target_expression: str = Field(alias='targetExpression')
    time_grains: List[TimeGrain] = Field(alias='timeGrains')
    time_dimension: str = Field(alias='timeDimension')
    dimensions: List[str]

    filters: Optional[List[Filter]]
    meta: Optional[Dict[str, str]]
    label: Optional[str]
    description: Optional[str]

    class Config:
        orm_mode = True
        allow_population_by_field_name = True


class MetricUpdate(BaseModel):
    id: int
    name: Optional[str]
    ds_dataset_id: Optional[int] = Field(alias='datasetId')
    aggregation_type: Optional[str] = Field(alias='aggregationType')
    target_expression: Optional[str] = Field(alias='targetExpression')
    time_grains: Optional[List[TimeGrain]] = Field(alias='timeGrains')
    time_dimension: Optional[str] = Field(alias='timeDimension')
    dimensions: Optional[List[str]]

    filters: Optional[List[Filter]]
    meta: Optional[Dict[str, str]]
    label: Optional[str]
    description: Optional[str]

    class Config:
        allow_population_by_field_name = True



import functools
import logging


def parametrized(decorator):
    @functools.wraps(decorator)
    def decorator_maker(*args, **kwargs):
        def decorator_wrapper(func):
            return decorator(func, *args, **kwargs)

        return decorator_wrapper

    return decorator_maker


@parametrized
def deprecated_without_replacement(func, version):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logging.warning(f"Function {func.__name__} will be deprecated in version {version}.")
        return func(*args, **kwargs)

    return wrapper


@parametrized
def deprecated_with_replacement(func, version, replacement):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        logging.warning(f"Function {func.__name__} will be deprecated in version {version}, use {replacement} instead.")
        return func(*args, **kwargs)

    return wrapper



from datetime import datetime
from optparse import Option
from typing import Any, List, Optional, Dict

from pydantic import Field
from pydantic_yaml import YamlModel, YamlStrEnum


class DatasetSourceType(YamlStrEnum):
    CSV = 'CSV'
    RASGO = 'RASGO'
    TABLE = 'TABLE'
    DATAFRAME = 'DATAFRAME'


class RasgoVersion(YamlStrEnum):
    V20 = '2.0'


class BaseV2(YamlModel):
    schema_version: RasgoVersion = RasgoVersion.V20

    class Config:
        allow_population_by_field_name = True
        orm_mode = True


class OfflineOperation(BaseV2):
    resource_key: Optional[str]
    operation_name: str
    operation_args: Optional[Dict[str, Any]]
    transform_name: Optional[str]
    dependencies: Optional[List[str]]
    fqtn: str


class OfflineOperationSet(BaseV2):
    resource_key: Optional[str]
    operations: Optional[List[OfflineOperation]]
    dependencies: Optional[List[str]]
    sql: Optional[str]


class OfflineInsight(BaseV2):
    name: Optional[str]
    transform_name: str
    transform_arguments: Optional[Dict[str, Any]]


class OfflineFilter(BaseV2):
    column_name: str
    operator: str
    comparison_value: str


class OfflineMetric(BaseV2):
    name: str
    aggregation_type: str
    target_expression: str
    time_grains: List[str]
    time_dimension: str
    dimensions: List[str]

    filters: Optional[List[OfflineFilter]]
    meta: Optional[Dict[str, str]]
    label: Optional[str]
    description: Optional[str]

    recent_values: Optional[List]
    recent_time_grain: Optional[str]
    recent_period_start: Optional[datetime]
    recent_period_end: Optional[datetime]


class OfflineColumn(BaseV2):
    name: str
    data_type: str


class OfflineDataset(BaseV2):
    name: str
    resource_key: str
    description: Optional[str]
    status: Optional[str]
    source_type: Optional[DatasetSourceType] = DatasetSourceType.RASGO
    organization_id: int

    attributes: Optional[Dict[str, str]]
    tags: Optional[List[str]]

    dw_operation_set: Optional[OfflineOperationSet]
    columns: Optional[List[OfflineColumn]]
    metrics: Optional[List[OfflineMetric]]
    insights: Optional[List[OfflineInsight]]

    fqtn: Optional[str]



"""
dbt Primitives
"""
import os
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union, Optional

from pyrasgo import schemas


class DbtSourceParent:
    """
    Represents a dbt source
    """

    def __init__(
        self,
        name: str,
        config_args: Dict[str, Any] = None,
    ):
        """
        Initializes a DbtSource class
        """
        self.name = name
        self.config_args = config_args


class DbtSource:
    """
    Represents a dbt source
    """

    def __init__(
        self,
        name: str,
        fqtn: str,
        parent: DbtSourceParent,
        columns_schema: List[Tuple[str, str]] = None,
        config_args: Dict[str, Any] = None,
    ):
        """
        Initializes a DbtSource class
        """
        self.name = name
        self.parent = parent
        self.fqtn = fqtn
        self.columns_schema = columns_schema
        self.config_args = config_args


class DbtModel:
    """
    Represents a dbt model
    """

    def __init__(
        self,
        name: str,
        sql_definition: str,
        fqtn: str,
        columns_schema: List[Tuple[str, str]] = None,
        config_args: Dict[str, Any] = None,
        metrics: Optional[List[schemas.Metric]] = None,
    ):
        """
        Initializes a DbtModel class
        """
        self.name = name
        self.sql_definition = sql_definition
        self.fqtn = fqtn
        self.columns_schema = columns_schema
        self.config_args = config_args
        self.metrics = metrics

    # def save_file(
    #     self,
    #     models_directory: Union[os.PathLike, str],
    #     config_args: Dict[str, Any] = None,
    #     ref_tables: Dict[str, str] = None,
    # ) -> None:
    #     """
    #     Writes this dbt model to a file in the models dir
    #     """
    #     from pyrasgo.utils.dbt import save_model_file

    #     if self.config_args:
    #         if config_args:
    #             config_args = self.config_args.update(config_args)
    #         else:
    #             config_args = self.config_args
    #     return save_model_file(
    #         sql_definition=self.sql_definition,
    #         output_directory=Path(models_directory),
    #         file_name=self.name,
    #         schema=self.schema,
    #         config_args=config_args,
    #         ref_tables=ref_tables,
    #         include_schema=True if self.schema else False,
    #     )


class DbtProject:
    """
    Represents a dbt project
    """

    def __init__(
        self,
        name: str,
        project_directory: Union[os.PathLike, str],
        models_directory: Union[os.PathLike, str],
        models: List[DbtModel],
        sources: List[DbtSource],
        model_args: Dict[str, Any] = None,
    ):
        """
        Initializes a DbtProject class
        """
        self.name = name
        self.project_directory = Path(project_directory)
        self.models_directory = Path(models_directory)
        self.models = models
        self.sources = sources
        self.model_args = model_args

    def save_files(self, verbose: bool = False):
        """
        Writes all dbt models in this project as files to the project dir
        """
        from pyrasgo.utils.dbt import write_dbt_files

        return write_dbt_files(
            project_name=self.name,
            project_directory=self.project_directory,
            models_directory=self.models_directory,
            models=self.models,
            sources=self.sources,
            model_args=self.model_args,
            verbose=verbose,
        )
